@@ -23,13 +23,8 @@ package com.ibm.java.lang.management.internal;
 
 import java.io.IOException;
 import java.lang.management.GarbageCollectorMXBean;
-import java.lang.management.LockInfo;
-import java.lang.management.MemoryNotificationInfo;
 import java.lang.management.MemoryType;
-import java.lang.management.MemoryUsage;
-import java.lang.management.MonitorInfo;
 import java.lang.management.RuntimeMXBean;
-import java.lang.management.ThreadInfo;
 import java.lang.reflect.Array;
 import java.lang.reflect.InvocationTargetException;
 import java.lang.reflect.Method;
@@ -41,8 +36,6 @@ import java.util.Map;
 import java.util.Map.Entry;
 import java.util.Properties;
 import java.util.Set;
-import java.util.concurrent.ConcurrentHashMap;
-import java.util.function.Function;
 
 import javax.management.InstanceNotFoundException;
 import javax.management.MBeanServerConnection;
