@@ -1,5 +1,5 @@
 /*******************************************************************************
- * Copyright (c) 2016, 2018 IBM Corp. and others
+ * Copyright (c) 2016, 2020 IBM Corp. and others
  *
  * This program and the accompanying materials are made available under
  * the terms of the Eclipse Public License 2.0 which accompanies this
@@ -100,6 +100,10 @@ final class ArrayVarHandle extends VarHandle {
 		super(arrayType.getComponentType(), new Class<?>[] {arrayType, int.class}, populateMHs(arrayType), 0);
 	}
 
+	public MethodType accessModeTypeUncached(AccessMode accessMode) {
+		throw OpenJDKCompileStub.OpenJDKCompileStubThrowError();
+	}
+
 	/**
 	 * Type specific methods used by array element VarHandle methods.
 	 */
